from pieraknet.packets.packet import Packet


class IncompatibleProtocol(Packet):
    packet_id = 0x19
    packet_type = 'incompatible_protocol'

    protocol_version: int = None
    magic: bytes = None
    server_guid: int = None

    def encode_payload(self):
        self.write_byte(self.protocol_version)
        self.write_magic(self.magic)
        self.write_long(self.server_guid)



from pieraknet.packets.packet import Packet


class OfflinePong(Packet):
    packet_id = 0x1c
    packet_type = 'offline_pong'

    client_timestamp: int = None
    server_guid: int = None
    magic: bytes = None
    server_name: str = None

    def encode_payload(self):
        self.write_long(self.client_timestamp)
        self.write_long(self.server_guid)

        if not isinstance(self.magic, bytes):
            self.magic = str(self.magic)
            self.magic = self.magic.encode()


        self.write_magic(self.magic)
        if not isinstance(self.server_name, bytes):
            self.server_name: bytes = self.server_name.encode('utf-8')
        self.write_string(self.server_name)




import socket
import random
import sys
import time
import logging
from pieraknet.protocol_info import ProtocolInfo
from pieraknet.packets.offline_ping import OfflinePing
from pieraknet.handlers.offline_ping import OfflinePingHandler
from pieraknet.packets.open_connection_request_1 import OpenConnectionRequest1
from pieraknet.handlers.open_connection_request_1 import OpenConnectionRequest1Handler
from pieraknet.packets.open_connection_request_2 import OpenConnectionRequest2
from pieraknet.handlers.open_connection_request_2 import OpenConnectionRequest2Handler


class ConnectionNotFound(Exception):
    pass


class Server:
    def __init__(self, hostname='0.0.0.0', port=19132, logger=logging.getLogger(__name__)):
        self.logger = logger
        self.start_time = round(time.time(), 2)
        self.hostname = hostname
        self.port = port
        self.ipv = 4
        self.name = ''
        self.protocol_version = 11
        self.guid = random.randint(0, sys.maxsize - 1)
        self.connections = []
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.SOL_UDP)
        self.start_time = time.time()
        self.maxsize = 4096
        self.magic = b'\x00\xff\xff\x00\xfe\xfe\xfe\xfe\xfd\xfd\xfd\xfd\x12\x34\x56\x78'
        self.running = False
        self.timeout = 15
        self.logger.info('Server initialized.')

    def get_time_ms(self):
        return round(time.time() - self.start_time, 4)

    def send(self, data, address: tuple):
        if not (isinstance(data, bytes)):
            data = str(data)
            data = data.encode()
        self.socket.sendto(data, address)

    def get_connection(self, address):
        for connection in self.connections:
            if connection.address == address:
                return connection
        raise ConnectionNotFound()

    def add_connection(self, connection):
        self.connections.append(connection)

    def start(self):
        self.running = True
        self.socket.bind((self.hostname, self.port))
        self.logger.info(f"Server started ({str(self.get_time_ms())}s).")
        while self.running:
            time.sleep(1 / 20)
            try:
                data, client = self.socket.recvfrom(self.maxsize)
            except OSError:
                pass
            else:
                if data[0] in [ProtocolInfo.OFFLINE_PING, ProtocolInfo.OFFLINE_PING_OPEN_CONNECTIONS]:
                    packet: OfflinePing = OfflinePing(data)
                    OfflinePingHandler.handle(packet, self, client)
                elif data[0] == ProtocolInfo.OPEN_CONNECTION_REQUEST_1:
                    packet: OpenConnectionRequest1 = OpenConnectionRequest1(data)
                    OpenConnectionRequest1Handler.handle(packet, self, client)
                elif data[0] == ProtocolInfo.OPEN_CONNECTION_REQUEST_2:
                    packet: OpenConnectionRequest2 = OpenConnectionRequest2(data)
                    OpenConnectionRequest2Handler.handle(packet, self, client)
                elif ProtocolInfo.FRAME_SET_0 <= data[0] <= ProtocolInfo.FRAME_SET_F:
                    connection = self.get_connection(client)
                    connection.handle(data)

    def stop(self):
        self.running = False
        self.socket.close()
        self.logger.info('Server stopped.')


if __name__ == '__main__':
    server = Server()
    server.name = 'MCPE;PieMC Server;589;1.20.0;2;20;13253860892328930865;Powered by PieMC;Survival;1;19132;19133;'
    try:
        server.start()
    except KeyboardInterrupt:
        server.logger.info('Stopping...')
        server.stop()



#
#
# //--------\\    [----------]   ||--------]   ||\      /||    ||----------]
# ||        ||         ||        ||            ||\\    //||    ||
# ||        //         ||        ||======|     || \\  // ||    ||
# ||-------//          ||        ||            ||  \\//  ||    ||
# ||                   ||        ||            ||   —–   ||    ||
# ||              [----------]   ||--------]   ||        ||    ||----------]
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# @author PieMC Team
# @link http://www.PieMC-Dev.github.io/
#
#
#

import struct
from io import BytesIO

class UnsupportedIPVersion(Exception):
    pass


class EOSError(Exception):
    pass


class BuffError(Exception):
    pass


class Buffer(BytesIO):
    def feos(self):
        if len(self.getvalue()[self.tell():]) == 0:
            return True
        else:
            return False

    def read_packet_id(self):  # Read Packet ID
        return self.read_byte()

    def write_packet_id(self, data):
        self.write_byte(str(data))

    def read_byte(self):
        return struct.unpack('B', self.read(1))[0]

    def write_byte(self, data):
        if not isinstance(data, bytes):
            data = str(data).encode()
        self.write(struct.pack('B', int(data)))
        
    def read_bits(self, num_bits):
        byte_data = self.read((num_bits + 7) // 8)
        bits = []
        for byte in byte_data:
            for i in range(7, -1, -1):
                bits.append((byte >> i) & 1)
        return bits[:num_bits]
    
    def write_bits(self, bit_array):
        num_bits = len(bit_array)
        num_bytes = (num_bits + 7) // 8
        byte_data = bytearray(num_bytes)
        for i, bit in enumerate(bit_array):
            byte_index = i // 8
            bit_index = 7 - (i % 8)
            if bit:
                byte_data[byte_index] |= (1 << bit_index)
        self.write(byte_data)
        
    def read_ubyte(self):
        return struct.unpack('<B', self.read(1))[0]

    def write_ubyte(self, data):
        if not isinstance(data, bytes):
            data = data.encode('utf-8')
        self.write(struct.pack('<B', data))

    def read_short(self):
        shrt = self.read(2)
        print(shrt)
        return struct.unpack('>h', shrt)[0]

    def write_short(self, data):
        self.write(struct.pack('>h', data))

    def read_unsigned_short(self):
        ushrt = self.read(2)
        return struct.unpack('>H', ushrt)[0]

    def write_unsigned_short(self, data):
        self.write(struct.pack('>H', data))

    def read_magic(self):
        return self.read(16)

    def write_magic(self, data=b'00ffff00fefefefefdfdfdfd12345678'):
        if not isinstance(data, bytes):
            data = data.encode('utf-8')
        self.write(data)

    def read_long(self):
        return struct.unpack('>q', self.read(8))[0]

    def write_long(self, data):
        self.write(struct.pack('>q', data))

    def read_ulong(self):
        return struct.unpack('>Q', self.read(8))[0]

    def write_ulong(self, data):
        self.write(struct.pack('>Q', data))

    def read_int(self):
        dat = self.read(4)
        print(dat)
        return struct.unpack(">i", dat)[0]

    def write_int(self, data):
        self.write(struct.pack('>i', data))

    def read_uint(self):
        return struct.unpack(">I", self.read(4))[0]

    def write_uint(self, data):
        self.write(struct.pack('>I', data))

    def read_bool(self):
        return struct.unpack('?', self.read(1))[0]

    def write_bool(self, data):
        self.write(struct.pack('?', data))

    def read_uint24le(self):
        uint24le = self.read(3) + b'\x00'
        return struct.unpack("<I", uint24le)[0]

    def write_uint24le(self, data):
        self.write(struct.pack("<I", data)[:3])

    def read_string(self):
        length = self.read_short()
        string = self.read(length).decode('ascii')
        return string

    def write_string(self, data):
        self.write_short(len(data))
        if not isinstance(data, bytes):
            data = data.encode('ascii')
        self.write(data)

    def read_address(self):
        ipv = self.read_byte()
        if ipv == 4:
            hostname_parts = []
            for part in range(4):
                hostname_parts.append(str(~self.read_byte() & 0xff))
            hostname = ".".join(hostname_parts)
            port = self.read_unsigned_short()
            return hostname, port
        else:
            raise UnsupportedIPVersion('IP version is not 4')

    def write_address(self, address: tuple):
        self.write_byte(4)
        hostname_parts: list = address[0].split('.')
        for part in hostname_parts:
            self.write_byte(~int(part) & 0xff)
        self.write_unsigned_short(address[1])

    def read_var_int(self):
        value: int = 0
        for i in range(0, 35, 7):
            number = self.read_ubyte()
            value |= ((number & 0x7f) << i)
            if (number & 0x80) == 0:
                return value
        raise BuffError("VarInt is too big")

    def write_var_int(self, value: int) -> None:
        value &= 0xffffffff
        for i in range(0, 5):
            to_write: int = value & 0x7f
            value >>= 7
            if value != 0:
                self.write_ubyte(to_write | 0x80)
            else:
                self.write_ubyte(to_write)
                break



from pieraknet.packets.packet import Packet
from pieraknet.buffer import Buffer
from pieraknet.protocol_info import ProtocolInfo


class Acknowledgement(Packet):
    def __init__(self, data: bytes = b''):
        super().__init__(data)
        self.sequence_numbers: list[int] = []

    def decode_payload(self) -> None:
        self.sequence_numbers.clear()
        count: int = self.read_short()
        for i in range(0, count):
            single: bool = self.read_bool()
            if not single:
                index: int = self.read_uint24le()
                end_index: int = self.read_uint24le()
                while index <= end_index:
                    self.sequence_numbers.append(index)
                    index += 1
            else:
                self.sequence_numbers.append(self.read_uint24le())

    def encode_payload(self) -> None:
        self.sequence_numbers.sort()
        temp_buffer: Buffer = Buffer()
        count: int = 0
        if len(self.sequence_numbers) > 0:
            start_index: int = self.sequence_numbers[0]
            end_index: int = self.sequence_numbers[0]
            for pointer in range(1, len(self.sequence_numbers)):
                current_index: int = self.sequence_numbers[pointer]
                diff: int = current_index - end_index
                if diff == 1:
                    end_index: int = current_index
                elif diff > 1:
                    if start_index == end_index:
                        temp_buffer.write_bool(True)
                        temp_buffer.write_uint24le(start_index)
                        start_index = end_index = current_index
                    else:
                        temp_buffer.write_bool(False)
                        temp_buffer.write_uint24le(start_index)
                        temp_buffer.write_uint24le(end_index)
                        start_index = end_index = current_index
                    count += 1
            if start_index == end_index:
                temp_buffer.write_bool(True)
                temp_buffer.write_uint24le(start_index)
            else:
                temp_buffer.write_bool(False)
                temp_buffer.write_uint24le(start_index)
                temp_buffer.write_uint24le(end_index)
            count += 1
            self.write_short(count)
            self.write(temp_buffer.getvalue())


class Nack(Acknowledgement):
    packet_id = ProtocolInfo.NACK


class Ack(Acknowledgement):
    packet_id = ProtocolInfo.NACK



from pieraknet.packets.packet import Packet
import zlib


class GamePacket(Packet):
    packet_id = 0xfe
    packet_type = 'game_packet'

    body: bytes = None

    def decode_payload(self, decompress=True):
        self.body = self.read()
        if decompress:
            self.body = self.decompress(self.body)

    def encode_payload(self, compress=True):
        if compress:
            self.write(self.compress(self.body))
        else:
            self.write(self.body)

    @staticmethod
    def decompress(data):
        return zlib.decompress(data, -zlib.MAX_WBITS, 1024 * 1024 * 8)

    @staticmethod
    def compress(data):
        compress = zlib.compressobj(1, zlib.DEFLATED, -zlib.MAX_WBITS)
        compressed_data = compress.compress(data)
        compressed_data += compress.flush()
        return compressed_data



from pieraknet.buffer import Buffer


class Packet(Buffer):
    packet_id: int = None
    packet_type: int = None

    def encode_header(self, data):
        self.write_packet_id(data)

    def decode_header(self):
        return self.read_packet_id()

    def encode(self):
        self.encode_header(self.packet_id)
        if hasattr(self, 'encode_payload'):
            self.encode_payload()

    def decode(self):
        self.decode_header()
        if hasattr(self, 'decode_payload'):
            self.decode_payload()



from pieraknet.packets.packet import Packet


class ConnectionRequestAccepted(Packet):
    packet_id = 0x10
    packet_type = 'connection_request_accepted'

    client_address: tuple = None
    system_index: int = None
    internal_ids: list[tuple] = None
    request_time: int = None
    accepted_time: int = None

    def encode_payload(self):
        self.write_address(self.client_address)
        self.write_short(self.system_index)
        for address in self.internal_ids:
            self.write_address(address)
        self.write_long(self.request_time)
        self.write_long(self.accepted_time)



from pieraknet.packets.packet import Packet


class OpenConnectionRequest1(Packet):
    packet_id = 0x05
    packet_type = 'open_connection_request_1'

    magic: bytes = None
    protocol_version: int = None
    mtu_size: int = None

    def decode_payload(self):
        self.magic = self.read_magic()
        self.protocol_version = int(self.read_byte())
        self.mtu_size = len(self.read()) + 46



from pieraknet.packets.open_connection_request_2 import OpenConnectionRequest2
from pieraknet.packets.open_connection_reply_2 import OpenConnectionReply2
from pieraknet.connection import Connection


class OpenConnectionRequest2Handler:
    @staticmethod
    def handle(packet: OpenConnectionRequest2, server, address: tuple):
        packet.decode()
        server.logger.debug("New Packet:")
        server.logger.debug(f"- Packet ID: {str(packet.getvalue()[0])}")
        server.logger.debug(f"- Packet Body: {str(packet.getvalue()[1:])}")
        server.logger.debug(f"- Packet Name: Open Connection Request 2")
        server.logger.debug(f"- MAGIC: {str(packet.magic)}")
        server.logger.debug(f"- Server Address: {str(packet.server_address)}")
        server.logger.debug(f"- MTU Size: {str(packet.mtu_size)}")
        server.logger.debug(f"- Client GUID: {str(packet.client_guid)}")

        new_packet = OpenConnectionReply2()
        new_packet.magic = packet.magic  # TODO: server.magic
        new_packet.server_guid = server.guid
        new_packet.client_address = address
        new_packet.mtu_size = packet.mtu_size
        new_packet.encode()

        server.send(new_packet.getvalue(), address)

        server.logger.debug("Sent Packet:")
        server.logger.debug(f"- Packet ID: {str(new_packet.getvalue()[0])}")
        server.logger.debug(f"- Packet Body: {str(new_packet.getvalue()[1:])}")
        server.logger.debug(f"- Packet Name: Open Connection Reply 2")
        server.logger.debug(f"- MAGIC: {str(new_packet.magic)}")
        server.logger.debug(f"- Server GUID: {str(new_packet.server_guid)}")
        server.logger.debug(f"- Client Address: {str(new_packet.client_address)}")
        server.logger.debug(f"- MTU Size: {str(new_packet.mtu_size)}")

        connection = Connection(address, server, packet.mtu_size, packet.client_guid)
        server.add_connection(connection)



from pieraknet.packets.packet import Packet


class NewIncomingConnection(Packet):
    packet_id = 0x13
    packet_type = 'new_incoming_connection'

    server_address: tuple = None
    internal_address: tuple = None # ('255.255.255.255', 19132)

    def encode_payload(self):
        self.write_address(self.server_address)
        self.write_address(self.internal_address)

    def decode_payload(self):
        self.server_address = self.read_address()
        self.internal_address = self.read_address()



from pieraknet.packets.packet import Packet


class OfflinePing(Packet):
    packet_id = 0x01
    packet_type = 'offline_ping'

    client_timestamp: int = None
    magic: bytes = None
    client_guid: int = None

    def decode_payload(self):
        self.client_timestamp = self.read_long()
        self.magic = self.read_magic()
        self.client_guid = self.read_long()



from pieraknet.packets.packet import Packet


class OpenConnectionReply2(Packet):
    packet_id = 0x08
    packet_type = 'open_connection_reply_2'

    magic: bytes = None
    server_guid: int = None
    client_address: tuple = None  # ('255.255.255.255', 19132)
    mtu_size: int = None
    encryption_enabled: bool = None

    def encode_payload(self):
        self.write_magic(self.magic)  # TODO: server.magic
        self.write_long(self.server_guid)
        self.write_address(self.client_address)
        self.write_short(self.mtu_size)
        self.write_bool(self.encryption_enabled)



from pieraknet.packets.packet import Packet


class OpenConnectionReply1(Packet):
    packet_id = 0x06
    packet_type = 'open_connection_reply_1'

    magic: bytes = None
    server_guid: int = None
    use_security: bool = None
    mtu_size: int = None

    def encode_payload(self):
        self.write_magic(self.magic)  # TODO: server.magic
        self.write_long(self.server_guid)
        self.write_bool(self.use_security)
        self.write_short(self.mtu_size)



from pieraknet.buffer import Buffer

class FrameSet:
    def __init__(self):
        self.sequence_number = 0
        self.frames = []

    def encode(self):
        buf = Buffer()
        buf.write_uint24le(self.sequence_number)

        for frame in self.frames:
            buf.write(frame.encode())

        return buf.getvalue()

    def get_size(self):
        return 3 + sum(frame.get_size() for frame in self.frames)

    def getvalue(self):
        return self.encode()

    def decode(self, data):
        buf = Buffer(data)
        self.sequence_number = buf.read_uint24le()

        while not buf.feos():
            frame = Frame()
            frame.decode(buf)
            self.frames.append(frame)

class Frame:
    def __init__(self):
        self.flags = 0
        self.reliability = 0
        self.fragmented = False
        self.length_bits = 0
        self.reliable_frame_index = 0
        self.sequenced_frame_index = 0
        self.ordered_frame_index = 0
        self.order_channel = 0
        self.compound_size = 0
        self.compound_id = 0
        self.index = 0
        self.body = b''

    def decode(self, buf):
        self.flags = buf.read_byte()
        self.reliability = self.flags >> 5
        self.fragmented = (self.flags & 0x10) != 0
        self.length_bits = buf.read_unsigned_short()

        if self.reliability == 0:
            self.reliable_frame_index = buf.read_uint24le()
        elif self.reliability == 1:  # Add handling for other reliability types
            self.sequenced_frame_index = buf.read_uint24le()
        elif self.reliability == 2:
            self.ordered_frame_index = buf.read_uint24le()
            self.order_channel = buf.read_byte()

        if self.fragmented:
            self.compound_size = buf.read_int()
            self.compound_id = buf.read_short()
            self.index = buf.read_int()

        num_bytes = (self.length_bits + 7) // 8
        self.body = buf.read_bits(self.length_bits)[:num_bytes]
        
    def encode(self):
        buf = Buffer()
        flags = (self.reliability << 5) | (0x10 if self.fragmented else 0)  # Flags: reliability + fragmentation
        body_bits = self.body + [0] * (8 - len(self.body) % 8)  # Pad bits to byte boundary

        buf.write_byte(flags)
        buf.write_unsigned_short(len(body_bits))  # Set body length in bits

        if self.reliability in [1, 2, 3]:
            buf.write_uint24le(self.reliable_frame_index)
        if self.reliability == 1:
            buf.write_uint24le(self.sequenced_frame_index)
        if self.reliability == 2:
            buf.write_uint24le(self.ordered_frame_index)
            buf.write_byte(self.order_channel)
        if self.fragmented:
            buf.write_int(self.compound_size)
            buf.write_short(self.compound_id)
            buf.write_int(self.index)

        num_bytes = (len(body_bits) + 7) // 8
        byte_data = bytearray()
        for i in range(0, len(body_bits), 8):
            byte = 0
            for j in range(8):
                if i + j < len(body_bits):
                    byte |= body_bits[i + j] << (7 - j)
            byte_data.append(byte)
        buf.write(byte_data)
        
        return buf.getvalue()

    def get_size(self):
        return len(self.encode())

    def getvalue(self):
        return self.encode()


import time
from pieraknet.packets.online_ping import OnlinePing
from pieraknet.packets.online_pong import OnlinePong


class OnlinePingHandler:
    @staticmethod
    def handle(packet: OnlinePing, server, connection):
        packet.decode()
        new_packet = OnlinePong()
        new_packet.client_timestamp = packet.client_timestamp
        new_packet.server_timestamp = int(time.time())
        new_packet.encode()
        return new_packet.getvalue()



from pieraknet.packets.packet import Packet


class OpenConnectionRequest2(Packet):
    packet_id = 0x07
    packet_type = 'open_connection_request_2'

    magic: bytes = None
    server_address: tuple = None  # ('255.255.255.255', 19132
    mtu_size: int = None
    client_guid: int = None

    def decode_payload(self):
        self.magic = self.read_magic()
        self.server_address = self.read_address()
        self.mtu_size = self.read_short()
        self.client_guid = self.read_long()



import time
from pieraknet.packets.frame_set import FrameSet, Frame
from pieraknet.protocol_info import ProtocolInfo
from pieraknet.packets.acknowledgement import Ack, Nack
from pieraknet.handlers.connection_request import ConnectionRequestHandler
from pieraknet.handlers.online_ping import OnlinePingHandler
from pieraknet.packets.online_ping import OnlinePing
from pieraknet.packets.new_incoming_connection import NewIncomingConnection
from pieraknet.packets.disconnect import Disconnect
from pieraknet.buffer import Buffer

class Connection:
    def __init__(self, address, server, mtu_size, guid):
        self.address = address
        self.mtu_size = mtu_size
        self.server = server
        self.guid = guid
        self.connected = False
        self.recovery_queue = {}
        self.ack_queue = []
        self.nack_queue = []
        self.fragmented_packets = {}
        self.compound_id = 0
        self.client_sequence_numbers = []
        # self.server_sequence_numbers = []
        self.client_sequence_number = 0
        self.server_sequence_number = 0
        self.queue = FrameSet()
        self.server_reliable_frame_index = 0
        self.client_reliable_frame_index = 0
        self.channel_index = [0] * 32
        self.last_receive_time = time.time()

    def update(self):
        if (time.time() - self.last_receive_time) >= self.server.timeout:
            self.disconnect()
        self.send_ack_queue()
        self.send_nack_queue()
        self.send_queue()

    def send_data(self, data: bytes):
        self.server.send(data, self.address)

    def handle(self, data):
        self.last_receive_time = time.time()
        print("New Packet: ", data)  # Log the received packet data
        if data[0] == ProtocolInfo.ACK:
            self.handle_ack(data)
        elif data[0] == ProtocolInfo.NACK:
            self.handle_nack(data)
        elif ProtocolInfo.FRAME_SET_0 <= data[0] <= ProtocolInfo.FRAME_SET_F:
            self.handle_frame_set(data)  # Pass the raw binary data to the method
            print("Frame Set handled.")  # Log that the frame set was handled


    def handle_ack(self, data: bytes):
        print("Handling ACK packet...")
        packet = Ack(data)
        packet.decode()
        for sequence_number in packet.sequence_numbers:
            if sequence_number in self.recovery_queue:
                del self.recovery_queue[sequence_number]

    def handle_nack(self, data: bytes):
        print("Handling NACK packet...")
        packet = Nack(data)
        packet.decode()
        for sequence_number in packet.sequence_numbers:
            if sequence_number in self.recovery_queue:
                lost_packet = self.recovery_queue[sequence_number]
                lost_packet.sequence_number = self.server_sequence_number
                self.server_sequence_number += 1
                lost_packet.encode()
                self.send_data(lost_packet.data)
                del self.recovery_queue[sequence_number]

    def handle_frame_set(self, data):
        print("Handling Frame Set...")
        buf = Buffer(data)  # Create a Buffer instance from the received data
        frame_set = FrameSet()
        frame_set.decode(data)  # Pass the Buffer instance to the decode method
        if frame_set.sequence_number not in self.client_sequence_numbers:
            if frame_set.sequence_number in self.nack_queue:
                self.nack_queue.remove(frame_set.sequence_number)
            self.client_sequence_numbers.append(frame_set.sequence_number)
            self.ack_queue.append(frame_set.sequence_number)
            hole_size = frame_set.sequence_number - self.client_sequence_number
            if hole_size > 0:
                for sequence_number in range(self.client_sequence_number + 1, hole_size):
                    if sequence_number not in self.client_sequence_numbers:
                        self.nack_queue.append(sequence_number)
            self.client_sequence_number = frame_set.sequence_number
            for frame in frame_set.frames:
                if not (2 <= frame.reliability <= 7 and frame.reliability != 5):
                    self.handle_frame(frame)
                else:
                    hole_size = frame.reliable_frame_index - self.client_reliable_frame_index
                    if hole_size == 0:
                        self.handle_frame(frame)
                        self.client_reliable_frame_index += 1


    def handle_fragmented_frame(self, packet):
        print("Handling Fragmented Frame...")
        if packet.compound_id not in self.fragmented_packets:
            self.fragmented_packets[packet.compound_id] = {packet.index: packet}
        else:
            self.fragmented_packets[packet.compound_id][packet.index] = packet
        if len(self.fragmented_packets[packet.compound_id]) == packet.compound_size:
            new_frame = Frame()
            new_frame.body = b''
            for i in range(0, packet.compound_size):
                new_frame.body += self.fragmented_packets[packet.compound_id][i].body
            del self.fragmented_packets[packet.compound_id]
            self.handle_frame(new_frame)

    def handle_frame(self, packet):
        print("Handling Frame...")
        if packet.fragmented:
            self.handle_fragmented_frame(packet)
        else:
            if not self.connected:
                if packet.body[0] == ProtocolInfo.CONNECTION_REQUEST:
                    new_frame = Frame()
                    new_frame.reliability = 0
                    new_frame.body = ConnectionRequestHandler.handle(packet.body, self.server, self)
                    self.add_to_queue(new_frame)
                elif packet.body[0] == ProtocolInfo.NEW_INCOMING_CONNECTION:
                    packet = NewIncomingConnection(packet.body)
                    packet.decode()
                    if packet.server_address[1] == self.server.port:
                        self.connected = True
                        if hasattr(self.server, "interface"):
                            if hasattr(self.server.interface, "on_new_incoming_connection"):
                                self.server.interface.on_new_incoming_connection(self)
            elif packet.body[0] == ProtocolInfo.ONLINE_PING:
                new_frame = Frame()
                new_frame.reliability = 0
                new_frame.body = OnlinePingHandler.handle(OnlinePing(packet.body), self, self.server)
                self.add_to_queue(new_frame, False)
            elif packet.body[0] == ProtocolInfo.DISCONNECT:
                self.disconnect()
            elif packet.body[0] == ProtocolInfo.GAME_PACKET:
                if hasattr(self.server, "interface"):
                    if hasattr(self.server.interface, "on_game_packet"):
                        self.server.interface.on_game_packet(packet, self)
            else:
                if hasattr(self.server, "interface"):
                    if hasattr(self.server.interface, "on_unknown_packet"):
                        self.server.interface.on_unknown_packet(packet, self)

    def send_queue(self):
        print("Sending Queue...")
        if len(self.queue.frames) > 0:
            self.queue.sequence_number = self.server_sequence_number
            self.server_sequence_number += 1
            self.recovery_queue[self.queue.sequence_number] = self.queue
            self.queue.encode()
            self.send_data(self.queue.getvalue())
            self.queue = FrameSet()

    def add_to_queue(self, packet: Frame, is_immediate=True):
        if 2 <= packet.reliability <= 7 and packet.reliability != 5:
            packet.reliable_frame_index = self.server_reliable_frame_index
            self.server_reliable_frame_index += 1
            if packet.reliability == 3:
                packet.ordered_frame_index = self.channel_index[packet.order_channel]
                self.channel_index[packet.order_channel] += 1
        if packet.get_size() > self.mtu_size:
            fragmented_body = []
            for i in range(0, len(packet.body), self.mtu_size):
                fragmented_body.append(packet.body[i:i + self.mtu_size])
            for index, body in enumerate(fragmented_body):
                new_packet = Frame()
                new_packet.fragmented = True
                new_packet.reliability = packet.reliability
                new_packet.compound_id = self.compound_id
                new_packet.compound_size = len(fragmented_body)
                new_packet.index = index
                new_packet.body = body
                if index != 0:
                    new_packet.reliable_frame_index = self.server_reliable_frame_index
                    self.server_reliable_frame_index += 1
                if new_packet.reliability == 3:
                    new_packet.ordered_frame_index = packet.ordered_frame_index
                    new_packet.order_channel = packet.order_channel
                if is_immediate:
                    self.queue.frames.append(new_packet)
                    self.send_queue()
                else:
                    frame_size: int = new_packet.get_size()
                    queue_size: int = self.queue.get_size()
                    if frame_size + queue_size >= self.mtu_size:
                        self.send_queue()
                    self.queue.frames.append(new_packet)
            self.compound_id += 1
        else:
            if is_immediate:
                self.queue.frames.append(packet)
                self.send_queue()
            else:
                frame_size: int = packet.get_size()
                queue_size: int = self.queue.get_size()
                if frame_size + queue_size >= self.mtu_size:
                    self.send_queue()
                self.queue.frames.append(packet)

    def send_ack_queue(self):
        if len(self.ack_queue) > 0:
            packet = Ack()
            packet.sequence_numbers = self.ack_queue
            self.ack_queue = []
            packet.encode()
            self.send_data(packet.getvalue())

    def send_nack_queue(self):
        if len(self.nack_queue) > 0:
            packet = Nack()
            packet.sequence_numbers = self.nack_queue
            self.nack_queue = []
            packet.encode()
            self.send_data(packet.getvalue())

    def disconnect(self):
        print("Disconnecting...")
        new_frame = Frame()
        new_frame.reliability = 0
        disconnect_packet = Disconnect()
        disconnect_packet.encode()
        new_frame.body = disconnect_packet.getvalue()
        self.add_to_queue(new_frame)
        self.server.remove_connection(self.address)
        if hasattr(self.server, "interface"):
            if hasattr(self.server.interface, "on_disconnect"):
                self.server.interface.on_disconnect(self)


from pieraknet.packets.packet import Packet


class ConnectionRequest(Packet):
    packet_id = 0x09
    packet_type = 'connection_request'

    client_guid: int = None
    client_timestamp: int = None

    def decode_payload(self):
        self.client_guid = self.read_long()
        self.client_timestamp = self.read_long()

    def encode_payload(self):
        self.write_long(self.client_guid)
        self.write_long(self.client_timestamp)



from pieraknet.packets.offline_ping import OfflinePing
from pieraknet.packets.offline_pong import OfflinePong


class OfflinePingHandler:
    @staticmethod
    def handle(packet: OfflinePing, server, address: tuple):
        packet.decode()
        server.logger.debug("New Packet:")
        server.logger.debug(f"- Packet ID: {str(packet.getvalue()[0])}")
        server.logger.debug(f"- Packet Body: {str(packet.getvalue()[1:])}")
        server.logger.debug(f"- Packet Name: Offline Ping")
        server.logger.debug(f"- Client Timestamp: {str(packet.client_timestamp)}")
        server.logger.debug(f"- MAGIC: {str(packet.magic)}")
        server.logger.debug(f"- Client GUID: {str(packet.client_guid)}")
        new_packet = OfflinePong()
        new_packet.client_timestamp = packet.client_timestamp
        new_packet.server_guid = server.guid
        new_packet.magic = packet.magic  # TODO: server.magic
        new_packet.server_name = server.name
        new_packet.encode()
        server.send(new_packet.getvalue(), address)
        server.logger.debug("Sent Packet:")
        server.logger.debug(f"- Packet ID: {str(new_packet.getvalue()[0])}")
        server.logger.debug(f"- Packet Body: {str(new_packet.getvalue()[1:])}")
        server.logger.debug(f"- Packet Name: Offline Pong")
        server.logger.debug(f"- Client Timestamp: {str(new_packet.client_timestamp)}")
        server.logger.debug(f"- Server GUID: {str(new_packet.server_guid)}")
        server.logger.debug(f"- MAGIC: {str(new_packet.magic)}")
        server.logger.debug(f"- Server Name: {str(new_packet.server_name)}")
